#
# Racc version 0.9.5
#
#    Copyright(c) 1999 Minero Aoki
#    aamine@dp.u-netsurf.ne.jp
#

require 'must'
require 'extmod'
require 'parser'
require 'scanner'


$RACCPARSER_DEBUG = false

class Racc

  Version = '0.9.5'

  attr :code
  attr :classname

  attr :parser
  attr :interf
  attr :prectable
  attr :ruletable
  attr :statetable
  attr :formatter

  attr :logic
  attr :srconf
  attr :rrconf

  attr :dsrc, true

  
  attr :debug,    true
  attr :d_prec,   true
  attr :d_rule,   true
  attr :d_token,  true
  attr :d_state,  true
  attr :d_reduce, true
  attr :d_shift,  true


  def compile( str, fname = '' )
    reset
    parse( str, fname )
    init_rule
    init_state
    resolve
    return output
  end


  def reset
    @ruletable  = RuleTable.new( self )
    @interf     = BuildInterface.new( self )
    @code       = {}
    @parser     = RaccParser.new( self )
    @statetable = LALRstatTable.new( self )
    @formatter  = RaccFormatter.new( self )
  end

  def parse( str, fname = '' )
    str.must String
    @parser.parse( str, fname )
    @classname = @parser.classname
  end

  def init_rule
    @ruletable.do_initialize( @parser.start )
  end

  def init_state
    @statetable.do_initialize
  end

  def resolve
    @statetable.resolve
  end

  def output
    @formatter.source
  end


  def initialize
    @logic  = []
    @rrconf = []
    @srconf = []
  end



  class RaccScanner < Scanner

    def reset( str, rac )
      super( str )
      @toktbl = rac.ruletable.tokentable
      @code   = rac.code
    end
    
    # pattern

    COMMENT   = /\A\#[^\n\r]*/o
    BEGIN_C   = /\A\/\*[^\n\r\*\/]*/o
    ATOM      = /\A[a-zA-Z_]\w*|\A\$end/o
    CODEBLOCK = /\A(?:\n|\r\n|\r)\-\-\-\-+/o


    def scan
      while true do
        unless @scan.rest? then
          @spipe.push false ; @vpipe.push false 
          @spipe.push false ; @vpipe.push false
          break
        end

        @scan.skip SPC
        unless @scan.skip COMMENT then
          if @scan.skip BEGIN_C then
            scan_comment
            next
          end
        end

        if temp = @scan.scan( CODEBLOCK ) then
          @scan.unscan
          @scan.skip EOL ; @lineno += 1
          scan_usercode
          @scan.clear
          next
        end

        if @scan.skip EOL then
          @lineno += 1
          next
        end

        if atm = @scan.scan( ATOM ) then
          scan_atom( atm )
          break  #next
        end

        case fch = @scan.getch
        when '"', "'"
          @spipe.push :STRING
          @vpipe.push scan_string( fch )
        when '{'
          @spipe.push :ACTION
          @vpipe.push scan_action
        else
          @spipe.push fch
          @vpipe.push fch
        end

        break
      end

      debug_report if @debug
    end



    private


    def scan_atom( cur )
      sret = :TOKEN
      vret = @toktbl.get_token( cur.intern )

      case cur
      when 'end'      then sret = :XEND
      when 'token'    then sret = :XTOKEN
      when 'right'    then sret = :RIGHT
      when 'left'     then sret = :LEFT
      when 'nonassoc' then sret = :NONASSOC
      when 'preclow'  then sret = :PRECLOW
      when 'prechigh' then sret = :PRECHIGH
      when 'start'    then sret = :START
      when 'class'    then sret = :CLASS
      when 'rule'     then sret = :RULE
      when 'file'     then sret = :XFILE
      when '$end'     then vret = @toktbl.get_token( Parser::Anchor )
      end

      @spipe.push sret
      @vpipe.push vret
    end


  # BEGIN_C   = /\A\/\*[^\n\r\*\/]*/o
    COM_ENT   = /\A[^\n\r*]+/o
    COM_ENT2  = /\A\*+[^*\/\r\n]/o
    END_C     = /\A\*+\//o

    def scan_comment
      while @scan.rest? do
        if    @scan.skip COM_ENT
        elsif @scan.skip COM_ENT2
        elsif @scan.skip EOL      then @lineno += 1
        elsif @scan.skip END_C    then return
        else
          scan_bug! 'in comment, no exp match'
        end
      end
      scan_error! 'find unterminated comment'
    end


    SKIP         = /\A[^\'\"\`\{\}\/\#\r\n]+/o
    COMMENT_CONT = /\A[^\r\n]*/o

    def scan_action
      ret  = ''
      nest = 0
      while @scan.rest? do
        if temp = @scan.scan( SKIP ) then
          ret << temp
        end
        if temp = @scan.scan( EOL ) then
          ret << temp
          @lineno += 1
          next
        end

        case ch = @scan.getch
        when '{'
          nest += 1
          ret << ch

        when '}'
          nest -= 1
          if nest < 0 then
            break
          end
          ret << ch

        when "'", '"', '`'
          ret << ch << scan_string( ch ) << ch

        when '/'
          if SPC === ret[-1,1] then
            if @scan.peep(1) != '=' then
              ret << ch << scan_string( ch ) << ch
              next
            end
          end
          ret << ch

        when '#'
          ret << ch << @scan.scan( COMMENT_CONT ) << @scan.scan( EOL )
          @lineno += 1

        else
          bug! 'must not huppen'
        end
      end

      return ret
    end


    INCLUDE    = /\A\s+\=/o
    BEGINBLOCK = /\A\-\-\-\-+/o

    def scan_usercode
      bname = nil
      ret = ''

      while @scan.rest? do
        unless line = @scan.scan( LINE ) then
          bug! 'in scan_rest, not match'
        end

        if BEGINBLOCK === line then
          if bname then
            @code.store( bname, ret )
            ret = ''
          end
          bname, rest = get_blockname( $' )

          if INCLUDE === rest then
            arr = get_filename( $' )
            load_files( arr, ret, bname )
          end
        else
          ret << line
        end
        @lineno += 1
      end
            
      if bname then
        @code.store( bname, ret )
      end
    end


    BLOCKNAME = /\A[ \t]*(\w+)/o

    def get_blockname( str )
      unless BLOCKNAME === str then
        scan_error! 'missing block name'
      end
      return $1.downcase, $'
    end


    def get_filename( str )
      str.strip!
      return str.split( /[ \t]/o )
    end


    def load_files( arr, ret, bname )
      arr.each do |fname|
        unless File.exist? fname then
          raise NameError, "#{bname} file '#{fname}' does not exist"
        end
        unless File.file? fname then
          raise NameError, "#{bname} file '#{fname}' is not file"
        end
        f = File.open( fname )
        ret << f.read
        f.close

        ret << "\n"  # for safety
      end
    end
      

  end   # RaccScanner



  class RaccParser < Parser

    attr :start
    attr :classname


    def initialize( rac )
      @racc   = (rac.must Racc)
      @code   = @racc.code
      @interf = @racc.interf

      @scanner = RaccScanner.new

      @scanner.debug = @__debug__ = $RACCPARSER_DEBUG
    end


    def parse( str, fname )
      @scanner.reset( str, @racc )
      @sbuf = @scanner.spipe
      @vbuf = @scanner.vpipe

      @filename = (fname.must String)
      @scanner.filename = @filename

      do_parse
    end

    def next_token
      if @sbuf.size == 0 then @scanner.scan end
      @sbuf.shift
    end

    def next_value
      if @vbuf.size == 0 then @scanner.scan end
      @vbuf.shift
    end

    def peep_token
      if @sbuf.size == 0 then @scanner.scan end
      @sbuf[0]
    end

    def on_error( etok, state, sstack, vstack )
      raise ParseError, <<MES
  parse error:#{@filename}:#{@scanner.lineno}: unexpected token '#{etok}'
MES
    end


##### racc generated code begin ###
LR_action_table = [
# state 0
Shift,
# state 1
Shift,
# state 2
Shift,
# state 3
:__reduce_with_rule_0,
# state 4
{
:XTOKEN => :__reduce_with_rule_2,
:START => :__reduce_with_rule_2,
:PRECHIGH => :__reduce_with_rule_2,
Default => :__reduce_with_rule_2,
:RULE => :__reduce_with_rule_2,
:XEND => :__reduce_with_rule_2,
:PRECLOW => :__reduce_with_rule_2
},
# state 5
Shift,
# state 6
Shift,
# state 7
:__reduce_with_rule_5,
# state 8
Shift,
# state 9
Shift,
# state 10
Shift,
# state 11
Shift,
# state 12
:__reduce_with_rule_1,
# state 13
:__reduce_with_rule_24,
# state 14
Shift,
# state 15
Shift,
# state 16
:__reduce_with_rule_25,
# state 17
:__reduce_with_rule_6,
# state 18
:__reduce_with_rule_11,
# state 19
{
:LEFT => :__reduce_with_rule_22,
:PRECHIGH => :__reduce_with_rule_22,
:TOKEN => :__reduce_with_rule_22,
:NONASSOC => :__reduce_with_rule_22,
Default => :__reduce_with_rule_22,
:STRING => :__reduce_with_rule_22,
:RIGHT => :__reduce_with_rule_22,
:PRECLOW => :__reduce_with_rule_22
},
# state 20
{
:LEFT => :__reduce_with_rule_22,
:PRECHIGH => :__reduce_with_rule_22,
:TOKEN => :__reduce_with_rule_22,
:NONASSOC => :__reduce_with_rule_22,
Default => :__reduce_with_rule_22,
:STRING => :__reduce_with_rule_22,
:RIGHT => :__reduce_with_rule_22,
:PRECLOW => :__reduce_with_rule_22
},
# state 21
{
:LEFT => :__reduce_with_rule_22,
:PRECHIGH => :__reduce_with_rule_22,
:TOKEN => :__reduce_with_rule_22,
:NONASSOC => :__reduce_with_rule_22,
Default => :__reduce_with_rule_22,
:STRING => :__reduce_with_rule_22,
:RIGHT => :__reduce_with_rule_22,
:PRECLOW => :__reduce_with_rule_22
},
# state 22
Shift,
# state 23
Shift,
# state 24
Shift,
# state 25
Shift,
# state 26
Shift,
# state 27
Shift,
# state 28
:__reduce_with_rule_4,
# state 29
:__reduce_with_rule_7,
# state 30
{
:LEFT => :__reduce_with_rule_13,
:TOKEN => Shift,
:PRECHIGH => :__reduce_with_rule_13,
:NONASSOC => :__reduce_with_rule_13,
Default => :__reduce_with_rule_13,
:STRING => Shift,
:RIGHT => :__reduce_with_rule_13,
:PRECLOW => :__reduce_with_rule_13
},
# state 31
{
:LEFT => :__reduce_with_rule_15,
:TOKEN => Shift,
:PRECHIGH => :__reduce_with_rule_15,
:NONASSOC => :__reduce_with_rule_15,
Default => :__reduce_with_rule_15,
:STRING => Shift,
:RIGHT => :__reduce_with_rule_15,
:PRECLOW => :__reduce_with_rule_15
},
# state 32
{
:LEFT => :__reduce_with_rule_14,
:TOKEN => Shift,
:PRECHIGH => :__reduce_with_rule_14,
:NONASSOC => :__reduce_with_rule_14,
Default => :__reduce_with_rule_14,
:STRING => Shift,
:RIGHT => :__reduce_with_rule_14,
:PRECLOW => :__reduce_with_rule_14
},
# state 33
:__reduce_with_rule_12,
# state 34
:__reduce_with_rule_9,
# state 35
Shift,
# state 36
:__reduce_with_rule_3,
# state 37
{
"|" => :__reduce_with_rule_22,
";" => :__reduce_with_rule_22,
:TOKEN => :__reduce_with_rule_22,
"=" => :__reduce_with_rule_22,
Default => :__reduce_with_rule_22,
:EOL => :__reduce_with_rule_22,
:STRING => :__reduce_with_rule_22,
:ACTION => :__reduce_with_rule_22
},
# state 38
{
"|" => :__reduce_with_rule_22,
";" => :__reduce_with_rule_22,
:TOKEN => :__reduce_with_rule_22,
"=" => :__reduce_with_rule_22,
Default => :__reduce_with_rule_22,
:EOL => :__reduce_with_rule_22,
:STRING => :__reduce_with_rule_22,
:ACTION => :__reduce_with_rule_22
},
# state 39
:__reduce_with_rule_19,
# state 40
:__reduce_with_rule_16,
# state 41
:__reduce_with_rule_18,
# state 42
:__reduce_with_rule_10,
# state 43
:__reduce_with_rule_8,
# state 44
:__reduce_with_rule_23,
# state 45
:__reduce_with_rule_17,
# state 46
{
"|" => :__reduce_with_rule_26,
";" => :__reduce_with_rule_26,
:TOKEN => Shift,
"=" => Shift,
Default => :__reduce_with_rule_26,
:STRING => Shift,
:EOL => :__reduce_with_rule_26,
:ACTION => :__reduce_with_rule_26
},
# state 47
{
"|" => :__reduce_with_rule_26,
";" => :__reduce_with_rule_26,
:TOKEN => Shift,
"=" => Shift,
Default => :__reduce_with_rule_26,
:STRING => Shift,
:EOL => :__reduce_with_rule_26,
:ACTION => :__reduce_with_rule_26
},
# state 48
{
"|" => :__reduce_with_rule_28,
";" => :__reduce_with_rule_28,
Default => :__reduce_with_rule_28,
:EOL => :__reduce_with_rule_28,
:ACTION => Shift
},
# state 49
Shift,
# state 50
{
"|" => :__reduce_with_rule_28,
";" => :__reduce_with_rule_28,
Default => :__reduce_with_rule_28,
:EOL => :__reduce_with_rule_28,
:ACTION => Shift
},
# state 51
:__reduce_with_rule_20,
# state 52
:__reduce_with_rule_29,
# state 53
:__reduce_with_rule_27,
# state 54
:__reduce_with_rule_21
]   # LR action table

LR_goto_table = [
# state 0
{
:xclass => 1,
:CLASS => 2
},
# state 1
{
Anchor => 3
},
# state 2
{
:TOKEN => 4
},
# state 3
nil,
# state 4
{
:clsentity => 5
},
# state 5
{
:XTOKEN => 6,
:xprec => 7,
:START => 8,
:PRECHIGH => 9,
:RULE => 10,
:XEND => 12,
:PRECLOW => 11
},
# state 6
{
:tokdefs => 14,
:TOKEN => 13,
:toksim => 15,
:STRING => 16
},
# state 7
nil,
# state 8
{
:TOKEN => 17
},
# state 9
{
:precline => 18,
:LEFT => 19,
:NONASSOC => 20,
:preclines => 22,
:RIGHT => 21
},
# state 10
{
:rules => 23,
:TOKEN => 24,
:ruleseg => 25
},
# state 11
{
:precline => 18,
:LEFT => 19,
:NONASSOC => 20,
:preclines => 26,
:RIGHT => 21
},
# state 12
nil,
# state 13
nil,
# state 14
{
:TOKEN => 13,
:toksim => 27,
:STRING => 16,
:XEND => 28
},
# state 15
{
:STRING => 29
},
# state 16
nil,
# state 17
nil,
# state 18
nil,
# state 19
{
:tokens => 30
},
# state 20
{
:tokens => 31
},
# state 21
{
:tokens => 32
},
# state 22
{
:precline => 33,
:LEFT => 19,
:NONASSOC => 20,
:RIGHT => 21,
:PRECLOW => 34
},
# state 23
{
:TOKEN => 24,
:XEND => 36,
:ruleseg => 35
},
# state 24
{
":" => 37
},
# state 25
{
"|" => 38,
";" => 39,
:rulesegterm => 40,
:EOL => 41
},
# state 26
{
:precline => 33,
:LEFT => 19,
:PRECHIGH => 42,
:NONASSOC => 20,
:RIGHT => 21
},
# state 27
{
:STRING => 43
},
# state 28
nil,
# state 29
nil,
# state 30
{
:TOKEN => 13,
:toksim => 44,
:STRING => 16
},
# state 31
{
:TOKEN => 13,
:toksim => 44,
:STRING => 16
},
# state 32
{
:TOKEN => 13,
:toksim => 44,
:STRING => 16
},
# state 33
nil,
# state 34
nil,
# state 35
{
"|" => 38,
";" => 39,
:rulesegterm => 45,
:EOL => 41
},
# state 36
nil,
# state 37
{
:tokens => 46
},
# state 38
{
:tokens => 47
},
# state 39
nil,
# state 40
nil,
# state 41
nil,
# state 42
nil,
# state 43
nil,
# state 44
nil,
# state 45
nil,
# state 46
{
:tempprec => 48,
:TOKEN => 13,
"=" => 49,
:toksim => 44,
:STRING => 16
},
# state 47
{
:tempprec => 50,
:TOKEN => 13,
"=" => 49,
:toksim => 44,
:STRING => 16
},
# state 48
{
:action => 51,
:ACTION => 52
},
# state 49
{
:TOKEN => 13,
:toksim => 53,
:STRING => 16
},
# state 50
{
:action => 54,
:ACTION => 52
},
# state 51
nil,
# state 52
nil,
# state 53
nil,
# state 54
nil
]   # LR goto table

DEBUG_PARSER = false

def __reduce_with_rule_0( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push Accept
 __state__[ -2, 2 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_1( vstack, sstack, __state__ )
 val = vstack[ -4, 4 ]
 sim = sstack[ -4, 4 ]
 vstack[ -4, 4 ] = []
 sstack[ -4, 4 ] = []
 sstack.push :xclass
 __state__[ -4, 4 ] = []
 result = val[0]
       @classname = val[1].to_s

 return result
end

def __reduce_with_rule_2( vstack, sstack, __state__ )
 val = []
 sim = []
 sstack.push :clsentity
 result = nil
 
 return result
end

def __reduce_with_rule_3( vstack, sstack, __state__ )
 val = vstack[ -4, 4 ]
 sim = sstack[ -4, 4 ]
 vstack[ -4, 4 ] = []
 sstack[ -4, 4 ] = []
 sstack.push :clsentity
 __state__[ -4, 4 ] = []
 result = val[0]
       @interf.end_register_rule

 return result
end

def __reduce_with_rule_4( vstack, sstack, __state__ )
 val = vstack[ -4, 4 ]
 sim = sstack[ -4, 4 ]
 vstack[ -4, 4 ] = []
 sstack[ -4, 4 ] = []
 sstack.push :clsentity
 __state__[ -4, 4 ] = []
 result = val[0]
       @interf.end_register_conv

 return result
end

def __reduce_with_rule_5( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :clsentity
 __state__[ -2, 2 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_6( vstack, sstack, __state__ )
 val = vstack[ -3, 3 ]
 sim = sstack[ -3, 3 ]
 vstack[ -3, 3 ] = []
 sstack[ -3, 3 ] = []
 sstack.push :clsentity
 __state__[ -3, 3 ] = []
 result = val[0]
       if @start then
        raise ParseError, "start rule defined twice"
      end
      @start = val[1]

 return result
end

def __reduce_with_rule_7( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :tokdefs
 __state__[ -2, 2 ] = []
 result = val[0]
       @interf.register_conv( val[0], val[1] )

 return result
end

def __reduce_with_rule_8( vstack, sstack, __state__ )
 val = vstack[ -3, 3 ]
 sim = sstack[ -3, 3 ]
 vstack[ -3, 3 ] = []
 sstack[ -3, 3 ] = []
 sstack.push :tokdefs
 __state__[ -3, 3 ] = []
 result = val[0]
       @interf.register_conv( val[1], val[2] )

 return result
end

def __reduce_with_rule_9( vstack, sstack, __state__ )
 val = vstack[ -3, 3 ]
 sim = sstack[ -3, 3 ]
 vstack[ -3, 3 ] = []
 sstack[ -3, 3 ] = []
 sstack.push :xprec
 __state__[ -3, 3 ] = []
 result = val[0]
       @interf.end_register_prec( true )

 return result
end

def __reduce_with_rule_10( vstack, sstack, __state__ )
 val = vstack[ -3, 3 ]
 sim = sstack[ -3, 3 ]
 vstack[ -3, 3 ] = []
 sstack[ -3, 3 ] = []
 sstack.push :xprec
 __state__[ -3, 3 ] = []
 result = val[0]
       @interf.end_register_prec( false )

 return result
end

def __reduce_with_rule_11( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :preclines
 __state__[ -1, 1 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_12( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :preclines
 __state__[ -2, 2 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_13( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :precline
 __state__[ -2, 2 ] = []
 result = val[0]
       @interf.register_prec( :Left, val[1] )

 return result
end

def __reduce_with_rule_14( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :precline
 __state__[ -2, 2 ] = []
 result = val[0]
       @interf.register_prec( :Right, val[1] )

 return result
end

def __reduce_with_rule_15( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :precline
 __state__[ -2, 2 ] = []
 result = val[0]
       @interf.register_prec( :Nonassoc, val[1] )

 return result
end

def __reduce_with_rule_16( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :rules
 __state__[ -2, 2 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_17( vstack, sstack, __state__ )
 val = vstack[ -3, 3 ]
 sim = sstack[ -3, 3 ]
 vstack[ -3, 3 ] = []
 sstack[ -3, 3 ] = []
 sstack.push :rules
 __state__[ -3, 3 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_18( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :rulesegterm
 __state__[ -1, 1 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_19( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :rulesegterm
 __state__[ -1, 1 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_20( vstack, sstack, __state__ )
 val = vstack[ -5, 5 ]
 sim = sstack[ -5, 5 ]
 vstack[ -5, 5 ] = []
 sstack[ -5, 5 ] = []
 sstack.push :ruleseg
 __state__[ -5, 5 ] = []
 result = val[0]
       @interf.register_rule( val[0], val[2], val[3], val[4] )

 return result
end

def __reduce_with_rule_21( vstack, sstack, __state__ )
 val = vstack[ -5, 5 ]
 sim = sstack[ -5, 5 ]
 vstack[ -5, 5 ] = []
 sstack[ -5, 5 ] = []
 sstack.push :ruleseg
 __state__[ -5, 5 ] = []
 result = val[0]
       @interf.register_rule( nil, val[2], val[3], val[4] )

 return result
end

def __reduce_with_rule_22( vstack, sstack, __state__ )
 val = []
 sim = []
 sstack.push :tokens
 result = nil
       result = []

 return result
end

def __reduce_with_rule_23( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :tokens
 __state__[ -2, 2 ] = []
 result = val[0]
       result.push val[1]

 return result
end

def __reduce_with_rule_24( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :toksim
 __state__[ -1, 1 ] = []
 result = val[0]
 
 return result
end

def __reduce_with_rule_25( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :toksim
 __state__[ -1, 1 ] = []
 result = val[0]
       result = @interf.get_token( eval '"' + val[0] + '"' )

 return result
end

def __reduce_with_rule_26( vstack, sstack, __state__ )
 val = []
 sim = []
 sstack.push :tempprec
 result = nil
 
 return result
end

def __reduce_with_rule_27( vstack, sstack, __state__ )
 val = vstack[ -2, 2 ]
 sim = sstack[ -2, 2 ]
 vstack[ -2, 2 ] = []
 sstack[ -2, 2 ] = []
 sstack.push :tempprec
 __state__[ -2, 2 ] = []
 result = val[0]
       result = val[1]

 return result
end

def __reduce_with_rule_28( vstack, sstack, __state__ )
 val = []
 sim = []
 sstack.push :action
 result = nil
       result = ''

 return result
end

def __reduce_with_rule_29( vstack, sstack, __state__ )
 val = vstack[ -1, 1 ]
 sim = sstack[ -1, 1 ]
 vstack[ -1, 1 ] = []
 sstack[ -1, 1 ] = []
 sstack.push :action
 __state__[ -1, 1 ] = []
 result = val[0]
 
 return result
end
##### racc generated code end #####

  end

  class BuildInterface

    def initialize( rac )
      @ruletable = rac.ruletable
      @tokentable = @ruletable.tokentable

      @precs = []

      @end_rule = false
      @end_conv = false
      @end_prec = false
    end

    
    def get_token( val )
      @tokentable.get_token( val )
    end
    
    def register_rule( simbol, rulearr, tempprec, actstr )
      @d_rule and
        puts "register: add: #{simbol} -> #{rulearr.join(' ')}"

      @end_rule and
        raise ParseError, "'rule' block is defined twice"

      if simbol then
        @pre = simbol
      else
        simbol = @pre
      end
      unless actstr then actstr = '' end

      @ruletable.register( simbol, rulearr, tempprec, actstr )
    end

    def end_register_rule
      @end_rule = true
    end


    def register_prec( atr, toks )
      @d_prec and
        puts "register: atr=#{atr.id2name}, toks=#{toks.join(' ')}"

      @end_prec and
        raise ParseError, "'prec' block is defined twice"

      toks.push atr
      @precs.push toks
    end

    def end_register_prec( rev )
      @end_prec = true

      top = @precs.size - 1
      @precs.each_with_index do |toks, idx|
        atr = toks.pop

        toks.each do |tok|
          tok.assoc = atr
          if rev then
            tok.prec = top - idx
          else
            tok.prec = idx
          end
        end
      end
    end


    def register_conv( tok, str )
      @end_conv and
        raise ParseError, "'token' block is defined twice"

      tok.conv = str
    end

    def end_register_conv
      @end_conv = true
    end

  end
    

  class RuleTable

    attr :tokentable


    def initialize( rac )
      @racc     = rac

      @d_token = rac.d_token
      @d_rule  = rac.d_rule
      @d_state = rac.d_state

      @tokentable = TokenTable.new

      @rules    = []
      @finished = false
      @hashval  = 4
    end


    def register( simbol, rulearr, tempprec, actstr )
      rule = Rule.new(
        simbol, rulearr, actstr,
        @rules.size + 1,         # ID
        @hashval,                # hash
        tempprec                 # prec
      )
      @rules.push rule

      @hashval += (rulearr.size + 2)
    end
   
    
    def do_initialize( start )

      ### add dammy rule

      start  = (start or @rules[0].simbol)
      dammy  = @tokentable.get_token( Parser::Dammy )
      anchor = @tokentable.get_token( Parser::Anchor )

      temp = Rule.new(
        dammy, [ start, anchor ], '',
        0,     # id
        0,     # hash
        nil    # prec token
      )
      @rules.unshift temp
      @rules.freeze


      ### cache

      @rules.each do |rule|
        rule.simbol.rules.push rule.ptrs(0)
      end

      @tokentable.each_token do |tok|
        tok.term = (tok.rules.size == 0)
      end

      @tokentable.each_token do |tok|
        tok.first
      end

      @rules.each do |rule|
        temp = nil

        rule.each_ptr do |ptr|
          tok = ptr.unref
          tok.locate.push ptr
          temp = tok if tok.term
        end

        rule.prec = temp unless rule.prec
      end
    end


    def []( x )
      @rules[x]
    end

    def each_rule
      @rules.each{|rl| yield rl }
    end

    def to_s
      "<Racc::RuleTable>"
    end


    def closure( ptrs )
      ptrs = orig_ptrs.uniq
      puts "closure: start: ptrs #{ptrs.join(' ')}" if @d_state

      temp = {}
      ptrs.each do |ptr|
        temp.store( ptr, true )

        tok = ptr.unref
        ptr.reduce? or tok.term or temp.update( tok.expand )
      end
      ret = temp.keys
      ret.sort!{|a,b| a.hash <=> b.hash }

      puts "closure: ret #{ret.join(' ')}" if @d_state
      return ret
    end

  end   # class RuleTable



  class Rule

    attr :action
    attr :simbol
    attr :ruleid
    attr :hashval
    attr :prec, true


    def []( idx ) @rulearr[idx] end

    def size ; @rulearr.size ; end

    def ptrs( idx ) @ptrs[idx] end

    def toks( idx ) @rulearr[idx] end

    def to_s() "<Rule:ID #{@ruleid}>" end

    def ==( other )
      (other.type == Rule) and (@ruleid == other.ruleid)
    end

    def hash() @hashval end

    def accept?()
      (tok = @rulearr[-1]) and tok.anchor?
    end

    def each_token
      @rulearr.each{|t| yield t }
    end

    def each_token_with_index
      @rulearr.each_with_index{|t, i| yield t, i }
    end

    def each_ptr
      pmax = @ptrs.size - 1
      pidx = 0
      while pidx < pmax do
        yield @ptrs[ pidx ]
        pidx += 1
      end
    end

    def initialize( tok, rlarr, actstr, rid, hval, tprec )
      @simbol  = tok
      @rulearr = rlarr
      @action  = actstr
      @ruleid  = rid
      @hashval = hval
      @prec    = tprec

      @ptrs = []
      rlarr.each_index do |idx|
        @ptrs[ idx ] = RulePointer.new( self, idx, rlarr[idx] )
      end

      # reduce pos
      s = rlarr.size
      @ptrs[ s ] = RulePointer.new( self, s, nil )
    end

  end   # class Rule



  class RulePointer

    attr :rule
    attr :ruleid
    attr :index
    attr :unref
    attr :data


    def initialize( rl, idx, tok )
      @rule   = rl
      @ruleid = rl.ruleid
      @index  = idx
      @unref  = tok

      @hashval = @rule.hash + @index
      @reduce  = tok.nil?
    end


    def to_s
      str = "(#{@rule.ruleid},#{@index}"
      if reduce? then
        str << '#)'
      else
        str << unref.to_s << ')'
      end
      return str
    end

    def inspect
      bug! 'ptr.inspect call'
    end

    def hash()     @hashval end
    def ==( ot )   @hashval == ot.hash end
    def eql?( ot ) @hashval == ot.hash end

    def reduce?()  @unref.nil? end
    def head?()    @index == 0 end

    def increment()
      (ret = @rule.ptrs(@index + 1)) or ptr_bug!
      return ret
    end

    def decrement()
      (ret = @rule.ptrs(@index - 1)) or ptr_bug!
      return ret
    end

    private
    
    def ptr_bug!
      bug! "pointer not exist: self: #{to_s}"
    end

  end   # class RulePointer



  class TokenTable

    def initialize
      @tokens  = {}

      # add system token data
      get_token( Parser::Default )
      get_token( Parser::Anchor )
      get_token( Parser::Dammy )
    end

    def get_token( val )

      ### for same value, Token exist only one

      unless (ret = @tokens.fetch( val )) then
        ret = Token.new( val )
        @tokens.store( val, ret )
      end

      return ret
    end

    def each_token
      @tokens.each_value{|tok| yield tok }
    end
    
  end   # class TokenTable



  class Token

    attr     :value
    attr     :rules
    attr     :locate
    attr     :term,                true
    attr     :nullp,               true
    attr     :assoc,               true
    attr     :prec,                true
    property :conv,  String, true, true


    Dammy   = Parser::Dammy
    Anchor  = Parser::Anchor
    Default = Parser::Default


    def initialize( tok )
      if tok.type == Token then tok = tok.value end
      @value = tok

      @valtype = @value.type
      @dammy   = (@value == Dammy)
      @anchor  = (@value == Anchor)
      @hashval = @value.hash

      @rules  = []
      @locate = []
      @term   = nil
      @first  = nil
      @conv   = nil
      @nullp  = nil
      @bfrom  = nil
    end


    def dammy?()  @dammy   end

    def anchor?() @anchor  end

    def hash()    @hashval end

    def to_s
      if    @value  ==  Dammy   then '$dammy'
      elsif Integer === @value  then @value.id2name
      elsif String  === @value  then @value.inspect
      elsif @value  ==  Anchor  then '$end'
      elsif @value  ==  Default then '$default'
      else
        bug! "token#uneval not match: val=#{@value}(#{@value.type})"
      end
    end
    alias inspect to_s


    def uneval
      if    @value  ==  Dammy   then 'Dammy'
      elsif Integer === @value  then ':' << @value.id2name
      elsif String  === @value  then @value.inspect
      elsif @value  ==  Anchor  then 'Anchor'
      elsif @value  ==  Default then 'Default'
      else
        bug! "token#uneval not match: val=#{@value}(#{@value.type})"
      end
    end


    def first
      @d_token and
        puts "get_first: start: token=#{self.to_s}"

      @first and return @first
      @first = {}

      ret = {}
      @nullp = false

      if @term then
        ret.store( self, true )
      else
        temp = {}
        @rules.each do |ptr|
          if ptr.reduce? then
            # reduce --> index 0 is null
            @nullp = true
          else
            temp.store( ptr.unref, true )
          end
        end

        temp.each_key do |tok|
          ret.update tok.first
        end
      end

      @first = ret
      @d_token and
        puts "get_first: for #{token}, ret #{ret.keys.join(' ')}"

      return ret
    end


    def expand
      @d_state and
        puts "expand: start: tok=#{self.to_s}"

      @bfrom and return @bfrom
      @bfrom = {}

      ret  = {}
      temp = {}
      @rules.each do |ptr|
        ret.store( ptr, true )
        tok = ptr.unref
        if not ptr.reduce? and not tok.term then
          temp.store( tok, true )
        end
      end
      temp.each_key do |tok|
        ret.update tok.expand
      end

      @bfrom = ret
      @d_state and
        puts "expand: ret #{ret.keys.join(' ')}"

      return ret
    end

  end   # Token



  class LALRstatTable

    def initialize( rac )
      @racc = rac
      @ruletable  = rac.ruletable
      @tokentable = @ruletable.tokentable
      @prectable  = rac.prectable

      @d_state  = rac.d_state
      @d_reduce = rac.d_reduce
      @d_shift  = rac.d_shift

      @stats = []
      @statcache = {}
    end
    

    def each_state
      @stats.each{|st| yield st }
    end

    def to_a() @stats end

    def to_s() "<LALRstatTable size #{@stats.size}>" end


    def do_initialize
      # add start closure
      add_new_state( [ @ruletable[0].ptrs(0) ] )

      cur = 0
      while cur < @stats.size do
        
        # state is added in develop_state
        develop_state( @stats[cur] )
        cur += 1
      end
    end


    def resolve
      @stats.each do |stat|
        act = stat.action
        if act.type == LookaheadAction then

          (@d_reduce or @d_shift) and
            puts "resolve state #{stat.stateid} -----------------"

          ### default

          de   = @tokentable.get_token( Parser::Default )
          rptr = stat.reduce_ptrs[0]
          act.store( de, ReduceAction.new(rptr) )

          ### resolve

          stat.resolve_rr
          stat.resolve_sr
        end
      end
    end


    private

    
    def develop_state( stat )
      @d_state and
        puts "develop_state: start\n#{stat.inspect}"

      stat.develop_seed.each do |tok, arr|
        @d_state and
          puts "init_stat: each: tok #{tok}, arr #{arr.join(' ')}"

        # 'arr's must be same order
        # (rule table order, upper to lower)
        arr.sort! do |a, b| a.hash <=> b.hash end

        unless (dest = @statcache[arr]) then
          # not registered yet
          dest = add_new_state( arr )
          
          @d_state and
            puts "init_stat: create dest.    ID #{dest.stateid}"
        else
          if @d_state then
            puts "init_stat: dest is cached. ID #{dest.stateid}"
            puts "init_stat: dest seed #{dest.seed.join(' ')}"
          end
        end

        stat.is_goto( tok, dest )
        dest.is_from( tok, stat )
      end
    end


    def add_new_state( seed )
      temp = LALRstat.new(
        closure( seed ),
        @stats.size,
        seed,
        @racc
      )
      @stats.push temp
      @statcache.store( seed, temp )

      return temp
    end


    def closure( orig_ptrs )
      ptrs = orig_ptrs.uniq
      @d_state and puts "closure: start: ptrs #{ptrs.join(' ')}"

      temp = {}
      ptrs.each do |ptr|
        temp.store( ptr, true )

        tok = ptr.unref
        ptr.reduce? or tok.term or temp.update( tok.expand )
      end
      ret = temp.keys
      ret.sort!{|a,b| a.hash <=> b.hash }

      @d_state and puts "closure: ret #{ret.join(' ')}"
      return ret
    end

  end   # LALRstatTable



  class LALRstat

    attr :stateid
    attr :ptrs

    attr :seed
    attr :develop_seed

    attr :action
    attr :reduce_ptrs
    attr :shift_ptrs

    attr :goto_table
    attr :from_table


    def initialize( arr, sid, sed, rac )
      @ptrs    = arr
      @stateid = sid
      @seed    = sed
      @racc    = rac
      @ruletable  = rac.ruletable
      @d_reduce   = rac.d_reduce
      @d_shift    = rac.d_shift

      @from_table = {}
      @goto_table = {}

      @lookahead = false

      @rrconf = []
      @srconf = []

      @reduce_ptrs  = []
      @shift_ptrs   = []
      @shift_toks   = []
      @develop_seed = {}

      @ptrs.each do |ptr|
        if ptr.reduce? then
          @reduce_ptrs.push ptr
        else
          @shift_ptrs.push ptr

          tok = ptr.unref
          if tok.term then
            @shift_toks.push tok
          end

          inc = ptr.increment

          if (arr = @develop_seed[ tok ]) then
            arr.push inc
          else
            @develop_seed.store( tok, [ inc ] )
          end
        end
      end
      @shift_toks.uniq!

      if @reduce_ptrs.size > 0 then
        if @ptrs.size == 1 then
          @action = ReduceAction.new( @reduce_ptrs[0] )
        else
          @action = LookaheadAction.new
        end
      else
        @action = ShiftAction.new
      end
    end


    def ==( oth ) @stateid == oth.stateid end

    def eql?( oth ) @seed == oth.seed end

    def stateq( oth ) @stateid == oth.stateid end

    def seedeq( oth ) @seed == oth end

    def hash() @stateid end

    def size() @ptrs.size end

    def to_s() "<LALR state #{@stateid}>" end

    def inspect
      "state #{@stateid}\n" + @ptrs.join("\n")
    end

    def each_ptr
      @ptrs.each{|ptr| yield ptr }
    end


    def is_goto( tok, dest )

      # check infinite recursion
      did = dest.stateid
      if (@stateid == did) and (size < 2) then
        rid = dest.ptrs[0].ruleid
        @racc.logic.push(
          "Infinite recursion: state #{did}, with rule #{rid}" )
      end

      # goto
      @goto_table.store( tok, dest )
    end

    
    def is_from( tok, stat )
      if (temp = @from_table[ tok ]) then
        temp.push stat
      else
        @from_table.store( tok, [ stat ] )
      end
    end


    def resolve_sr

      @shift_toks.each do |stok|
        unless (temp = @action.fetch( stok )) then
          # no conflict
          @action.store( stok, ShiftAction.new )
        else
          if temp.type == ReduceAction then
            # conflict on stok
            rtok = temp.value.prec
            ret = do_resolve_sr( stok, rtok )

            case ret
            when :Reduce
              # action is already set

            when :Shift
              # overwrite
              @action.store( stok, ShiftAction.new )

            when :Remove
              # remove action
              @action.delete( stok )

            when :CantResolve
              # shift to default
              @action.store( stok, ShiftAction.new )
              srconf( stok, temp.value )

            else
              bug! "do_resolve_sr return wrong val: #{ret}"
            end
          else
            # already set to shift
          end
        end
      end
    end


    def resolve_rr
      @reduce_ptrs.each do |curptr|
        @d_reduce and
          puts "resolve_rr: each: state #{@stateid}, #{curptr}"

        la_toks( curptr ).each_key do |tok|
          temp = @action.fetch( tok )
          if ReduceAction === temp then
            #
            # can't resolve R/R conflict (on tok),
            #   reduce with upper rule as default
            #
            rrconf( temp.value, curptr.rule, tok )
          else
            # no conflict
            @action.store( tok, ReduceAction.new( curptr ) )
          end
        end
      end
    end


    def lookahead( idlock )
      @d_reduce and puts "lookahead: state #{@stateid}"

      idlock[ @stateid ] = true         # recurcive lock
      @lookahead and return @lookahead  # cached

      ret  = {}

      temp = []
      @develop_seed.each_key do |tok|
        ret.update tok.first
        if tok.nullp then
          temp.push @goto_table.fetch( tok )
        end
      end
      @reduce_ptrs.each do |ptr|
        temp.concat backtrack( ptr )
      end
      temp.uniq!

      temp.each do |st|
        unless idlock[ st.stateid ] then
          ret.update st.lookahead( idlock )
        end
      end
      @lookahead = ret

      @d_reduce and puts "lookahead: lookahead #{ret.keys.join(' ')}"
      return ret
    end



    private

    
    def do_resolve_sr( stok, rtok )
      @d_shift and
        puts "resolve_sr: s/r conflict: rtok=#{rtok}, stok=#{stok}"

      unless rtok and rtok.prec then
        @d_shift and puts "resolve_sr: no prec for #{rtok}(R)"
        return :CantResolve
      end
      rprec = rtok.prec

      unless stok and stok.prec then
        @d_shift and puts "resolve_sr: no prec for #{stok}(S)"
        return :CantResolve
      end
      sprec = stok.prec

      if rprec == sprec then
        case rtok.assoc
        when :Left     then ret = :Reduce
        when :Right    then ret = :Shift
        when :Nonassoc then ret = :Remove  # 'rtok' raises ParseError
        else
          bug! "prec is not Left/Right/Nonassoc, #{rtok}"
        end
      else
        if rprec > sprec then
          ret = (if @reverse then :Shift else :Reduce end)
        else
          ret = (if @reverse then :Reduce else :Shift end)
        end
      end

      @d_shift and
        puts "resolve_sr: resolved as #{ret.id2name}"

      return ret
    end

    
    def la_toks( ptr )
      ret = {}

      backtrack( ptr ).each do |st|
        ret.update st.lookahead([])
      end

      return ret
    end


    def backtrack( ptr )
      @d_reduce and
        puts "backtrack: start: state #{@stateid}, ptr #{ptr}"

      sarr = [ self ]
      newstats = []

      until ptr.head? do
        ptr = ptr.decrement
        tok = ptr.unref

        sarr.each do |st|
          (bstats = st.from_table.fetch(tok)) or con_bug!( st, tok )
          newstats.concat bstats
        end
        newstats.uniq!

        sarr.replace newstats
        newstats.clear
      end

      sim = ptr.rule.simbol
      sarr.filter do |st| st.goto_table[ sim ] end

      @d_reduce and
        puts "backtrack: from #{@stateid} to #{sarr2s(sarr)}"

      return sarr
    end

    def sarr2s( sar ) sar.collect{|s| s.stateid}.join(' ') end

    def con_bug!( st, tok )
      bug! "from table void: state #{st.stateid} key #{tok}"
    end


    def rrconf( rule1, rule2, tok )
      @racc.rrconf.push RRconflict.new( @stateid, rule1, rule2, tok )
    end

    def srconf( stok, rrule )
      @racc.srconf.push SRconflict.new( @stateid, stok, rrule )
    end

  end   # LALRstat

  

  class LALRaction
    attr :value
  end

  class ShiftAction < LALRaction
    def initialize() @value = Parser::Shift end
    def to_s()       '<ShiftAction>'        end
  end

  class ReduceAction < LALRaction
    def initialize( ptr ) @value = ptr.rule end
    def to_s()            '<ReduceAction>'  end
  end

  class LookaheadAction < LALRaction
    def initialize()      @value = {}               end
    def each()            @value.each{|t| yield t } end
    def size()            @value.size               end
    def fetch( arg )      @value.fetch( arg )       end
    def store( arg, val ) @value.store( arg, val )  end
    def delete( key )     @value.delete( key )      end
    def to_s()            '<LookaheadAction>'       end
  end


  class ConflictData ; end

  class SRconflict < ConflictData

    attr :stateid
    attr :shift
    attr :reduce
  
    def initialize( sid, stok, rrule )
      @stateid = sid
      @shift   = stok
      @reduce  = rrule
    end
    
    def to_s
      str = sprintf(
        'state %d: S/R conflict rule %d reduce and shift %s',
        @stateid,
        @reduce.ruleid,
        @shift.to_s
      )
    end
  end

  class RRconflict < ConflictData
    
    attr :stateid
    attr :reduce1
    attr :reduce2
    attr :token
  
    def initialize( sid, rule1, rule2, tok )
      @stateid = sid
      @reduce1 = rule1
      @reduce2 = rule2
      @token   = tok
    end

    def to_s
      str =  sprintf(
        'state %d: R/R conflict with rule %d and %d on %s',
        @stateid,
        @reduce1.ruleid,
        @reduce2.ruleid,
        @token.to_s
      )
    end

  end


  class RaccFormatter


    def initialize( rac )
      @racc = rac
      @ruletable  = rac.ruletable
      @tokentable = @ruletable.tokentable
      @statetable = rac.statetable
      @dsrc       = rac.dsrc

      @temp = {}
    end



    def source
      str = "##### racc generated code begin ###\n"
      acttable_cat( str )
      gototable_cat( str )

      if @dsrc then
        str << "\nDEBUG_PARSER = true\n"
        tostable_cat( str )
      else
        str << "\nDEBUG_PARSER = false\n"
      end

      redumethods_cat( str )
      str << "##### racc generated code end #####\n"

      return str
    end


    def output_state
      str = "\n--------- State ---------\n\n"
      @statetable.each_state{|s| cat_state( str, s ) }
      str << "\n\n"

      return str
    end


    def output_rule
      str = "\n-------- Grammar --------\n\n"
      @ruletable.each_rule do |rl|
        cat_rule( str, rl ) if @racc.debug or rl.ruleid != 0
      end

      return str
    end


    def output_token
      str = "\n------- Token data -------\n\n"
      cat_toks( str )
      return str
    end



    private


    ########
    ######## .tab.rb
    ########


    def acttable_cat( str )
      str << 'LR_action_table = ['
      com = "\n"

      @statetable.each_state do |stat|
        act = stat.action

        str << com ; com = ",\n"
        str << "# state #{stat.stateid}\n"
        case act
        when LookaheadAction then laact_cat( str, act )
        when ShiftAction     then str << 'Shift'
        when ReduceAction
          str << ":__reduce_with_rule_#{act.value.ruleid}"
        else
          bug! "wrong action #{act.type} in state #{stat.stateid}"
        end
      end
      str << "\n]   # LR action table\n\n"
    end
      

    def gototable_cat( str )
      str << 'LR_goto_table = ['
      com = "\n"

      @statetable.each_state do |stat|
        str << com ; com = ",\n"
        str << "# state #{stat.stateid}\n"

        if stat.goto_table.size == 0 then
          str << 'nil'
        else
          str << '{'
          add = "\n"

          stat.goto_table.each do |tok, dest|
            str << "#{add}#{tok.conv or tok.uneval} => #{dest.stateid}"
            add = ",\n"
          end
          str << "\n}"
        end
      end
      str << "\n]   # LR goto table\n"
    end


    def redumethods_cat( str )
      @ruletable.each_rule do |rl|
        if rl.size == 0 then
          singleredu_cat( str, rl )
        else
          multiredu_cat( str, rl )
        end
      end
    end


    def multiredu_cat( str, rl )
      le = (rl.size * -1).to_s << ', '
      le << rl.size.to_s

      str.concat <<SOURCE

def __reduce_with_rule_#{rl.ruleid}( vstack, sstack, __state__ )
 val = vstack[ #{le} ]
 sim = sstack[ #{le} ]
SOURCE

      if @dsrc then
        str << %| if @__debug__ then\n|
        idx = 0
        rl.each_token do |tok|
          str << %|  print " #{tok}(\#{val[#{idx}].inspect})"\n|
          idx += 1
        end
        str << %|  print " --> #{rl.simbol.to_s}\\n"\n end\n|
      end

      str.concat <<SOURCE
 vstack[ #{le} ] = []
 sstack[ #{le} ] = []
 sstack.push #{rl.accept? ? 'Accept' : rl.simbol.uneval}
 __state__[ #{le} ] = []
 result = val[0]
 #{rl.action}
 return result
end
SOURCE
    end


    def singleredu_cat( str, rl )
      str.concat <<SOURCE

def __reduce_with_rule_#{rl.ruleid}( vstack, sstack, __state__ )
 val = []
 sim = []
SOURCE

      if @dsrc then
        str.concat <<SOURCE
if @__debug__ then
 print '<none> --> ', "#{rl.simbol.to_s}\\n"
end
SOURCE
      end

      str.concat <<SOURCE
 sstack.push #{rl.accept? ? 'Accept' : rl.simbol.uneval}
 result = nil
 #{rl.action}
 return result
end
SOURCE
    end


    def laact_cat( str, act )
      if act.value.size == 0 then
        bug! "LA action size is 0"
      end

      str << '{'
      com = "\n"
      act.value.each do |tok, act|
        str << com ; com = ",\n"
        str << "#{tok.conv or tok.uneval} => "
        case act
        when ShiftAction  then str << 'Shift'
        when ReduceAction then str << ":__reduce_with_rule_#{act.value.ruleid}"
        end
      end
      str << "\n}"
      str
    end


    def tostable_cat( str )
      str << "\nTOKEN_TO_S_TABLE = {"
      com = "\n"
      @tokentable.each_token do |tok|
        str << com ; com = ",\n"
        str << "#{tok.conv or tok.uneval} => '#{tok}'"
      end
      str << "\n}\n"
    end


    #########
    #########  .output
    #########


    def cat_state( str, stat )
      str << "state #{stat.stateid}\n"

      stat.ptrs.each{|pt| cat_ptr( str, pt ) }
      str << "\n"

      act = stat.action
      case act
      when LookaheadAction then cat_laact( str, stat )
      when ShiftAction     then cat_shift( str, stat )
      when ReduceAction    then cat_reduce( str, stat )
      else
        bug! "cat_state not match: act=#{act}(#{act.type})"
      end
      str << "\n\n"
    end


    def cat_ptr( str, ptr )
      rule = ptr.rule

      str << "#{rule.ruleid.to_s.rjust(4)}) #{rule.simbol} :"

      rule.each_token_with_index do |tok, idx|
        str << ' _' if idx == ptr.index
        str << " #{tok}"
      end
      if ptr.reduce? then
        str << ' _'
      end
      str << "\n"
    end


    def cat_laact( str, stat )
      tstr = ''
      nstr = ''

      stat.action.each do |tok, act|
        case act
        when ShiftAction  then cat_termline( tstr, tok, stat.goto_table[tok] )
        when ReduceAction then cat_lareduce( str, tok, act )
        else
          bug! "cat_laact not match: act=#{act}(#{act.type})"
        end
      end
      str << tstr

      stat.goto_table.each do |tok, dest|
        com = "\n"
        unless tok.term then
          str << com ; com = ''
          cat_ntermline( str, tok, stat.goto_table[tok] )
        end
      end
    end


    def cat_shift( str, stat )
      act = stat.action
      gstr = ''
      ssize = 0
      gsize = 0

      stat.goto_table.each do |tok,stat|
        if tok.term then cat_termline( str, tok, stat )   ; ssize += 1
        else             cat_ntermline( gstr, tok, stat ) ; gsize += 1
        end
      end

      if gsize > 0 then
        str << "\n" if ssize > 0
        str << gstr
      end
    end


    def cat_termline( str, tok, stat )
      str << sprintf( "  %-12s  shift, and go to state %d\n", 
                      tok.to_s, stat.stateid )
    end


    def cat_ntermline( str, tok, stat )
      str << sprintf( "  %-12s  go to state %d\n", 
                      tok.to_s, stat.stateid )
    end


    def cat_reduce( str, stat )
      rule = stat.action.value
      if rule.accept? then str << "  accept\n"
      else                 str << "  reduce using rule #{rule.ruleid}\n"
      end
    end


    def cat_lareduce( str, tok, act )
      rule = act.value
      if rule.accept? then str << sprintf( "  %-12s  accept\n", tok.to_s )
      else                 str << sprintf( "  %-12s  reduce using rule %d\n",
                                           tok.to_s, rule.ruleid )
      end
    end


    def cat_toks( str )
      str << "**Terminals, with rules where they appear\n\n"
      nstr = "\n**Nonterminals, with rules where they appear\n\n"

      @tokentable.each_token do |tok|
        if tok.term then cat_termtok( str, tok ) unless tok.anchor?
        else             cat_ntermtok( nstr, tok ) unless tok.dammy?
        end
      end

      str << nstr
    end


    def cat_termtok( str, tok )
      str << "  #{tok}"
      if temp = tok.conv then str << "(#{temp})" end
      str << "\n"

      # locate
      str << "    on right: #{arr2strnz(tok.locate)}\n\n"
    end


    def cat_ntermtok( str, tok )
      str << "  #{tok}\n"

      # locate
      if tok.locate.size > 0 then
        str << "    on right: #{arr2strnz(tok.locate)}\n"
      end

      # rule
      str << "    on left : #{arr2strnz(tok.rules)}\n\n"
    end


    def cat_rule( str, rl )
      str << sprintf( 'rule %-5d %s:',
                      rl.ruleid, rl.simbol.to_s )
      rl.each_token{|tok| str << " #{tok}" }
      str << "\n\n"
    end

    
    def arr2strnz( arr )
      arr.each do |ptr|
        @temp.store( ptr.ruleid, true ) unless ptr.ruleid == 0
      end
      ret = @temp.keys.join(' ')
      @temp.clear

      return ret
    end

  end


end
